!> \~japanese
!! \brief 正二十面体格子の一つの格子点に対して, 物理場の微分(勾配, 発散, 回転)を評価するためのサブルーチンを提供するモジュール.
!! \~english
!! \brief 正二十面体格子の一つの格子点に対して, 物理場の微分(勾配, 発散, 回転)を評価するためのサブルーチンを提供するモジュール.
!!
!! <br><br>
!! \b Copyright (C) GFD Dennou Club, 2011-2012. All rights reserved. <br>
!! \b license ?? <br>
!! \author Yuta Kawai
!!
module FVM_HDiff_operator

  ! モジュール引用 ; Use statement
  !
   
  ! 種類型パラメータ
  ! Kind type parameter
  !
  use dc_types, only: DP  ! 倍精度実数型. Double precision. 
  
  ! 線形代数
  ! Linear algebra
  !
  use igmcore_linear_algebra, only: &
    & vec_length, vec_normarize, &
    & cross, dot

  ! 球面三角法
  ! Spherical trigonometry
  !
  use igmcore_spherical_trigonometry, only: &
    & geodesic_arc_length

  ! 宣言文 ; Declaration statement
  !
  implicit none
  private

  ! 公開手続き 
  ! Public procedure
  !
  public :: eval_div_operator, eval_rot_operator, eval_grad_operator

contains


!
!>
!!
!!
function eval_div_operator( &
  & val_GP0, val_CV, GP0, CV,                   &  ! (in)
  & CVArea, ic_radius, CV_num, val_dim, ret_dim &  ! (in)
  & ) result(val)

  ! 宣言文 ; Declaration statements
  !

  integer, intent(in) :: val_dim
  integer, intent(in) :: ret_dim
  integer, intent(in) :: CV_num
  real(DP), intent(in) :: val_GP0(val_dim)
  real(DP), intent(in) :: GP0(3)
  real(DP), intent(in) :: val_CV(CV_num, val_dim)
  real(DP), intent(in) :: CV(CV_num, val_dim)
  real(DP), intent(in) :: CVArea
  real(DP), intent(in) :: ic_radius
  real(DP) val(ret_dim)

  ! 実行文 ; Executable statements
  !

  val(:) =  div_operator( &
             & vec_CV= val_CV, pt_CV=CV, CVArea=CVArea, &  ! (in)
             & pt_size=CV_num, ic_radius=ic_radius )       ! (in)

end function eval_div_operator

!
!>
!!
!!
function eval_rot_operator( &
  & val_GP0, val_CV, GP0, CV,                   &  ! (in)
  & CVArea, ic_radius, CV_num, val_dim, ret_dim &  ! (in)
  & ) result(val)

  ! 宣言文 ; Declaration statements
  !
  integer, intent(in) :: val_dim
  integer, intent(in) :: ret_dim
  integer, intent(in) :: CV_num
  real(DP), intent(in) :: val_GP0(val_dim)
  real(DP), intent(in) :: GP0(3)
  real(DP), intent(in) :: val_CV(CV_num, val_dim)
  real(DP), intent(in) :: CV(CV_num, val_dim)
  real(DP), intent(in) :: CVArea
  real(DP), intent(in) :: ic_radius
  real(DP) val(ret_dim)

  ! 実行文 ; Executable statements
  !

  val(:) =  rot_operator( &
             & vec_CV= val_CV, pt_CV=CV, CVArea=CVArea, &  ! (in)
             & pt_size=CV_num, ic_radius=ic_radius )       ! (in)

end function eval_rot_operator

!
!>
!!
!!
function eval_grad_operator( &
  & val_GP0, val_CV, GP0, CV,                   &  ! (in)
  & CVArea, ic_radius, CV_num, val_dim, ret_dim &  ! (in)
  & ) result(val)

  ! 宣言文 ; Declaration statements
  !
  integer, intent(in) :: val_dim
  integer, intent(in) :: ret_dim
  integer, intent(in) :: CV_num
  real(DP), intent(in) :: val_GP0(val_dim)
  real(DP), intent(in) :: GP0(3)
  real(DP), intent(in) :: val_CV(CV_num, val_dim)
  real(DP), intent(in) :: CV(CV_num, val_dim)
  real(DP), intent(in) :: CVArea
  real(DP), intent(in) :: ic_radius
  real(DP) val(ret_dim)

  ! 実行文 ; Executable statements
  !

!  val(:) =  grad_operator( &
!             & q_GP0=val_GP0(1), q_CV= val_CV(:,1), pt_CV=CV, CVArea=CVArea, pt_size=CV_num, ic_radius=ic_radius )


  val(:) =  grad_operator2( &
             & q_GP0=val_GP0(1), q_CV= val_CV(:,1), GP=GP0, pt_CV=CV, &  ! (in)
             & CVArea=CVArea, pt_size=CV_num, ic_radius=ic_radius )      ! (in)

end function eval_grad_operator

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! 非公開手続き
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!
!> ベクトル場の水平発散を計算する. 
!!
!! @param[in] vec_CV    コントロールボリュームの頂点上のベクトル場の値を保持する配列. 
!! @param[in] pt_CV     コントロールボリュームの頂点座標を保持する配列. 
!! @param[in] CVArea    コントロールボリュームの面積. 
!! @param[in] pt_size   格子点が持つコントールボリュームの頂点数. 
!! @param[in] ic_radius 正二十面体格子を内包する球の半径.  
!! @return    val       ベクトル場の水平発散. 
!!
function div_operator( vec_CV, pt_CV, CVArea, pt_size, ic_radius ) result(val)

  ! 宣言文 ; Declaration statement
  !
  integer, intent(in) :: pt_size
  real(DP), intent(in) :: vec_CV(pt_size, 3)
  real(DP), intent(in) :: pt_CV(pt_size, 3)
  real(DP), intent(in) :: CVArea
  real(DP), intent(in) :: ic_radius
  real(DP) val

  ! 作業変数 
  ! Work variables
  !
  integer :: i
  real(DP) :: n(3)

  ! 実行文 ; Executable statement
  !

  val = 0.0d0

  do i=1, pt_size
    n(:) = cross( pt_CV(1+mod(i, pt_size), :), pt_CV(i,:) )
    call vec_normarize(n)

    val = val + geodesic_arc_length(ic_radius, pt_CV(1+mod(i, pt_size), :), pt_CV(i,:)) * &
      &   dot( 0.5d0 * ( vec_CV(1+mod(i, pt_size), :) + vec_CV(i,:) ), n(:) )
  end do

  val = val / CVArea

end function div_operator

!
!> 水平ベクトル場の回転の鉛直成分を計算する. 
!!
!! @param[in] vec_CV
!! @param[in] pt_CV
!! @param[in] CVArea
!! @param[in] pt_size
!! @param[in] ic_radius
!! @return    val
!!
function rot_operator( vec_CV, pt_CV, CVArea, pt_size, ic_radius ) result(val)
  
  ! 宣言文 ; Declaration statement
  !
  integer, intent(in) :: pt_size
  real(DP), intent(in) :: vec_CV(pt_size, 3)
  real(DP), intent(in) :: pt_CV(pt_size, 3)
  real(DP), intent(in) :: CVArea
  real(DP), intent(in) :: ic_radius
  real(DP) val

  ! 作業変数
  ! Work variable
  !
  integer :: i
  real(DP) :: m(3)

  ! 実行文 ; Executable statement
  !
  val = 0.0d0

  do i=1, pt_size
    m(:) = pt_CV(1+mod(i, pt_size), :) - pt_CV(i,:) 
    call vec_normarize(m)

    val = val + geodesic_arc_length(ic_radius, pt_CV(1+mod(i, pt_size), :), pt_CV(i,:)) * &
      &   dot( 0.5d0 * ( vec_CV(1+mod(i, pt_size), :) + vec_CV(i,:) ), m(:) )
  end do

  val = val / CVArea

end function rot_operator

!
!> 2 次元のスカラー場の水平勾配を計算する. 
!!
!! @param[in] q_GP0
!! @param[in] q_CV
!! @param[in] pt_CV
!! @param[in] CVArea
!! @param[in] pt_size
!! @param[in] ic_radius
!! @return    val
!!
function grad_operator( q_GP0, q_CV, pt_CV, CVArea, pt_size, ic_radius ) result(val)

  ! 宣言文 ; Declaration statement
  !
  integer, intent(in) :: pt_size
  real(DP), intent(in) :: q_GP0
  real(DP), intent(in) :: q_CV(pt_size)
  real(DP), intent(in) :: pt_CV(pt_size, 3)
  real(DP), intent(in) :: CVArea
  real(DP), intent(in) :: ic_radius
  real(DP) val(3)

  ! 作業変数　
  ! Work variable
  !
  integer :: i
  real(DP) :: n(3)

  ! 実行文 ; Executable statement
  !

  val(:) = 0.0d0

  do i=1, pt_size
    n(:) = cross( pt_CV(1+mod(i, pt_size), :), pt_CV(i,:) )
    call vec_normarize(n)

    val(:) = val(:) + geodesic_arc_length(ic_radius, pt_CV(1+mod(i, pt_size), :), pt_CV(i,:)) * &
      &   ( 0.5d0 * ( q_CV(1+mod(i, pt_size)) + q_CV(i) ) - q_GP0 ) * n(:)
  end do

  val(:) = val(:) / CVArea

end function grad_operator

!
!>
!!
!! @param[in] q_GP0
!! @param[in] q_CV
!! @param[in] pt_CV
!! @param[in] CVArea
!! @param[in] pt_size
!! @param[in] ic_radius
!! @return    val
!!
function grad_operator2( q_GP0, q_CV, GP, pt_CV, CVArea, pt_size, ic_radius ) result(val)
  
  ! 宣言文  ; Declaration statement
  !
  integer, intent(in) :: pt_size
  real(DP), intent(in) :: q_GP0
  real(DP), intent(in) :: q_CV(pt_size)
  real(DP), intent(in) :: GP(3)
  real(DP), intent(in) :: pt_CV(pt_size, 3)
  real(DP), intent(in) :: CVArea
  real(DP), intent(in) :: ic_radius
  real(DP) val(3)

  ! 作業変数 
  ! Work variables
  !
  integer :: i
  real(DP) :: n(3)

  ! 実行文 ; Executable statement
  !

  val(:) = 0.0d0

  do i=1, pt_size
    n(:) = cross( pt_CV(1+mod(i, pt_size), :) - pt_CV(i,:), GP(:) )
    call vec_normarize(n)

    val(:) = val(:) + geodesic_arc_length(ic_radius, pt_CV(1+mod(i, pt_size), :), pt_CV(i,:)) * &
      &   ( 0.5d0 * ( q_CV(1+mod(i, pt_size)) + q_CV(i) ) - q_GP0 ) * n(:)
  end do

  val(:) = val(:) / CVArea

end function grad_operator2

end module FVM_HDiff_operator
